// Compiler options:-
// -DBALT     = make trees balanced

#include <iostream.h>
#include <fcntl.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <math.h>
#include <strings.h>
#include <string.h>

#include "Itemset.h"
#include "ListItemset.h"
#include "HashTree.h"
#include "Database.h"
#include "pardhp.h"

int **lsupcnt;
int mflag = 1;
int host_num=0;
#ifdef PARALLEL
#include "MC_lib.h"
#else
int MC_NUM_HOSTS=1;
int MC_PROC=1;
#endif
struct timeval tp;

#define MAXITER 30
#define MAXTRANSZ 100

int MINSUPPORT;
double MINSUP_PER = 0.25;
int NPROC = 1;
int NUM_INSERT;
int NUM_ACTUAL_INSERT;

HashTree *Candidate = NULL;
ListItemset *Largelist = NULL;

char infile[500], offset_file[500];
FILE *summary;
double ts, te, t_io_1, t_io_2;

int **start, **enda;
int **hash_pos;

int threshold=2;
int more;

int *hash_indx = NULL;
int tot_cand = 0;

extern transaction *data_set;

void parse_args(int argc, char **argv)
{
   extern char * optarg;
   int c;
   
   if (argc < 4){
      cout << "usage: -i<infile> -f<offset_file> -s<support>\n";
      cout << "-h<hash_function> -t<threshold>\n";
      exit(3);
   }
   else{      
      while ((c=getopt(argc,argv,"b:i:f:m:p:s:t:w:"))!=-1){
         switch(c){
         case 'b':
            DBASE_BUFFER_SIZE = atoi(optarg);
            break;
         case 'i':
            strcpy(infile,optarg);
            break; 
         case 'f':
            strcpy(offset_file, optarg);
            break;
#if defined PARALLEL
         case 'm':
            MC_NUM_HOSTS = atoi(optarg);
            break;
         case 'p':
            MC_PROC = atoi(optarg);
            break;
         case 'w':
            host_num = atoi(optarg);
            mflag = 0;
            break;
#endif
         case 's':
            MINSUP_PER = atof(optarg);
            break;
         case 't':
            threshold = atoi(optarg);
            break;
         }
      }
   }
}

#ifdef BALT
void form_hash_indx(int hash_function)
{
   int i, cnt;
   i=0;

   printf("HASH_FUNCTION = %d\n", hash_function);
   fprintf(summary,"HASH_FUNCTION = %d\n", hash_function);
   if (hash_function == 1){
      return;
   }

   while(i < DBASE_MAXITEM){
      for(cnt = 0; i < DBASE_MAXITEM && cnt < hash_function; i++)
         if (hash_indx[i] == 0){
            hash_indx[i] = cnt;
            cnt++;
         }
      for(cnt = hash_function-1;i < DBASE_MAXITEM && cnt >= 0; i++)
         if (hash_indx[i] == 0){
            hash_indx[i] = cnt;
            cnt--;
         }
   }
}
#endif

int init_subsets(int *starts, int *endas, int num_item, int k_item)
{
   int i;
   
   if (num_item < k_item) return 0;
   
   for (i=0; i < k_item; i++){
      starts[i] = i;
      endas[i] = num_item - k_item + 1 + i;
   }
   return 1;
}

int get_next_subset(int *starts, int *endas, int k_item)
{
   int i,j;
   
   for (i=k_item-1; i >= 0; i--){
      starts[i]++;
      if (starts[i] < endas[i]){
         for (j=i+1; j < k_item; j++)
            starts[j] = starts[j-1]+1;
         return 1;
      }
   }
   return 0;
}

ListElement * find_in_list(Itemset *item, int sz, ListElement *head, int pnum)
{
   for(;head; head = head->next()){
      Itemset *curr = head->item();
      for(int i=0; i < sz; i++){
         int it = item->item((start[pnum])[i]);
         if (curr->item(i) < it)
            break;
         else if (curr->item(i) > it)
            return NULL;
         else if (i==sz-1) return head;
      }
   }
   return NULL;
}


int not_prune(Itemset *curr, int k, ListElement *beg, int pnum)
{
   if (k+1 == 3){
      start[pnum][0] = 1;
      start[pnum][1] = 2;
      if ((beg = find_in_list(curr, k, beg, pnum)) == NULL) return 0;
   }
   else{
      int res = init_subsets(start[pnum], enda[pnum], curr->numitems(), k);
      start[pnum][k-2] = curr->numitems()-2;
      start[pnum][k-1] = curr->numitems()-1;
      while (res){
         if ((beg = find_in_list(curr, k, beg, pnum)) == NULL) return 0;
         res = get_next_subset(start[pnum], enda[pnum], k);
      }
   }
   return 1;
}

int choose(int n, int k)
{
   int i;
   int val = 1;

   if (k >= 0 && k <= n){
      for (i=n; i > n-k; i--)
         val *= i;
      for (i=2; i <= k; i++)
         val /= i;
   }
   return val;
}

int get_hash_function(int num, int k){
   int hash = (int)ceil(pow(num/threshold, 1.0/k));
   if (hash < 1) hash = 1;
   return hash;
}


void assign_lsupcnt_offset(HashTree *node, int &val)
{
   if (node->is_leaf()){
      ListItemset *list = node->list();
      if(list && list->first()){
         ListElement *iter = list->first();
         for(;iter;iter = iter->next()){
            iter->item()->set_sup(val++);
         }
      }
   }
   else{
      for(int i=0; i < node->hash_function(); i++)
         if (node->hash_table(i))
            assign_lsupcnt_offset(node->hash_table(i), val);
   } 
}

int apriori_gen(int k, int pnum){
   int tt;
   //long blk = (Largelist->numitems()+NPROC-1)/NPROC;
   long blk = Largelist->numitems();
   int lb = pnum*blk;
   int ub = (pnum+1)*blk;
   ListElement *L1iter = Largelist->node(lb);
   for (int i=lb; i < ub && L1iter; i++, L1iter = L1iter->next()){
      Itemset *temp = L1iter->item();
      ListElement *L2iter = L1iter->next();
      for(;L2iter; L2iter = L2iter->next()){
         Itemset *temp2 = L2iter->item();
         if (temp->compare(*temp2,k-2) < 0) break;
         else{
            int k1 = temp->item(k-2);
            int k2 = temp2->item(k-2); 
            if (k1 < k2){
               Itemset *it = new Itemset(k);
               it->set_numitems(k);
               for (int l=0; l < temp->numitems(); l++)
                  it->add_item(l,temp->item(l));
               it->add_item(k-1, k2);
               ListElement *beg = Largelist->first();
               if(k==2 || not_prune(it, k-1, beg, pnum)){
                  tt = Candidate->add_element(*it);
                  tot_cand++;
               }
            }
         }
      }
   }
#ifdef PARALLEL
   tt= 0;
   assign_lsupcnt_offset(Candidate,tt);
#endif
   return tt;
}


void increment(Itemset *trans, ListItemset *Clist, int pnum, char *tbitvec)
{
   if (Clist->first()){
      ListElement *head = Clist->first();
      for(;head; head = head->next()){
         Itemset *temp = head->item();
         if (temp->subsequence(tbitvec, trans->numitems())){
         //if (temp->subsequence(*trans)){
#ifdef PARALLEL
            lsupcnt[pnum][temp->sup()]++;
#else
            temp->incr_sup();
#endif
         }
      }
   }
}

void subset(Itemset *trans, int st, int en, int final,
            HashTree* node, int k, int level, int pnum, 
            char *tbitvec, int *vist, int hash_function)
{
   int i;

   (*vist)++;
   int myvist = *vist;

   if (node == Candidate 
       && node->is_leaf() && node->list())
      increment(trans, node->list(), pnum, tbitvec);
   else{
      for(i=st; i < en; i++){
#ifdef BALT
         int val = trans->item(i);
         int hashval = hash_indx[val];
         if (hashval == -1) continue;
#else
         int val = trans->item(i);
         if (hash_indx[val] == -1) continue;
         int hashval = node->hash(val);
#endif
         if ((hash_pos[pnum])[level*hash_function+hashval] != myvist){
            (hash_pos[pnum])[level*hash_function+hashval] = myvist;
            if (node->hash_table_exists() && node->hash_table(hashval)){
               if (node->hash_table(hashval)->is_leaf() &&
                   node->hash_table(hashval)->list())
                  increment(trans, node->hash_table(hashval)->list()
                            ,pnum, tbitvec);
               else if (en+1 <= final)
                  subset(trans, i+1, en+1, final,node->hash_table(hashval),
                         k, level+1, pnum, tbitvec, vist, hash_function);
            }
         }
      }
   }
}


void form_large(HashTree *node,int k,int &cnt, int *cntary)
{
   if (node->is_leaf()){
      ListItemset *list = node->list();
      if(list && list->first()){
         ListElement *iter = list->first();
         for(;iter;iter = iter->next()){
#ifdef PARALLEL
            iter->item()->set_sup(cntary[cnt++]);
#endif
            printf("SUP %d\n", iter->item()->sup());
	    fprintf(summary, "SUP %d\n", iter->item()->sup());
            if (iter->item()->sup() >= MINSUPPORT){               
               Largelist->sortedInsert(iter->item());
               for (int j=0; j < iter->item()->numitems(); j++)
               {
                   hash_indx[iter->item()->item(j)] = 0;
               }
            }      
         }
      }
   }
   else{
      for(int i=0; i < node->hash_function(); i++)
         if (node->hash_table(i))
            form_large(node->hash_table(i), k, cnt, cntary);
   } 
}



#ifdef PARALLEL
int *sum_large_counts(int *largecnt, int pnum, int lim, int iter)
{
   int i;
   int mc_res;

   int *Rshar = (int *) R_REGION->malloc_end;
   int *Tshar = (int *) getTaddr(Rshar);

   mc_lock_acquire(2);
   int tmp;
   if (R_REGION->valid == 0){
      R_REGION->valid = 1;
      T_REGION->valid = 1;
      for(i=0; i < lim; i++){
         Rshar[i] = largecnt[i];
         Tshar[i] = largecnt[i];
      }
   }
   else{
      for(i=0; i < lim; i++){
         tmp = Rshar[i]+largecnt[i];
         Rshar[i] = tmp;
         Tshar[i] = tmp;
      }
   }
   
   while(R_REGION->valid != 1);
   mc_lock_release(2);
   mc_sync();
   if (mflag){
      R_REGION->valid = 0;
      T_REGION->valid = 0;
   }
   
   while(R_REGION->valid != 0);
   return Rshar;
}
#endif

void alloc_mem_for_var(int pnum)
{
   start[pnum] = new int [MAXITER];
   enda[pnum] = new int [MAXITER];
}

void sum_counts(int *destary, int **srcary, int pnum, int ub)
{
   int i,j;
   
   for(i=1; i < NPROC; i++){
      for(j=0; j < ub; j++){
         destary[j] += srcary[i][j];
      }
   }
}

void *main_proc(void *)
{
   int vist, hash_function;
   double t1,t2,t3,t4;
   HashTree *oldCand=NULL;
   int k, i, j, pnum=0;
   Itemset *trans=NULL;
   int *buf, tid, numitem, idx;
   int *cnt_ary;
   int *offsets;
   FILE *off_file;
   int *file_offset;
 
  seconds(t1);
   seconds(t_io_1);
   file_offset = (int *) calloc(2, sizeof(int));
   off_file = fopen(offset_file, "r");
   for (i=0; i<2; i++)
     fscanf(off_file,"%d\n", &(file_offset[i]));

   int max_trans_sz = Database_readfrom(infile, lsupcnt[pnum], offsets, pnum, file_offset[0], file_offset[1] - file_offset[0]);

   trans = new Itemset(max_trans_sz);
   alloc_mem_for_var(pnum);
   char *trans_bitvec = new char[DBASE_MAXITEM];
   for (i=0; i < DBASE_MAXITEM; i++){
      trans_bitvec[i] = 0;
   }
   if (pnum == 0){
      //printf("DONE READING SUMING %d\n", get_host_num());
      fprintf(summary, "Database %s sup %f items %d minsup %d thr %d\n",
                 infile, MINSUP_PER, 
                 DBASE_NUM_TRANS,
                 MINSUPPORT, threshold);
  
      printf("Database %s sup %f items %d minsup %d thr %d\n",
                infile, MINSUP_PER,
                DBASE_NUM_TRANS,
                MINSUPPORT, threshold);
  
      hash_indx = new int [DBASE_MAXITEM];

#ifdef PARALLEL
      if (MC_NUM_HOSTS*MC_PROC > 1)
         cnt_ary = mc_reduce(lsupcnt[pnum], (DBASE_MAXITEM*(DBASE_MAXITEM-1))/2,
         mflag);
         //cnt_ary = sum_large_counts(lsupcnt[pnum], pnum,
         //                           (DBASE_MAXITEM*(DBASE_MAXITEM-1)), 1);
      else cnt_ary = lsupcnt[pnum];
#else
      cnt_ary = lsupcnt[pnum];
#endif
      //printf("DONE REDUCTION %d\n", get_host_num());
      for(i=0; i < DBASE_MAXITEM; i++)
         hash_indx[i] = -1;
      for(i=0; i < DBASE_MAXITEM-1; i++){
         idx = offsets[i]-i-1;
         for (j=i+1; j < DBASE_MAXITEM; j++){
            if (cnt_ary[idx+j] >= MINSUPPORT){
               hash_indx[i] = 0;
               hash_indx[j] = 0;
               Itemset *it = new Itemset(2);
               it->set_numitems(2);
               it->add_item(0,i);
               it->add_item(1,j);
               it->set_sup(cnt_ary[idx+j]);
               Largelist->append(*it);
               tot_cand ++;
            }
         }
      }
      
   
      NUM_INSERT = choose(Largelist->numitems(),2);
      printf("(2)it = %d ", Largelist->numitems());
      fprintf(summary, "(2)it = %d ", Largelist->numitems());
      hash_function = get_hash_function(NUM_INSERT,2);
      Candidate = new HashTree(0, hash_function, threshold);
#ifdef BALT
      more = (NUM_INSERT > 0);
      form_hash_indx(hash_function);
#endif
   }
   printf("NUM_INSERT = %d\n", NUM_INSERT);
   fprintf(summary,"NUM_INSERT = %d\n", NUM_INSERT);
   seconds(te);
   if (pnum == 0) printf (" %f ", te-ts);
   if (pnum == 0) fprintf (summary," %f ", te-ts);
#ifdef PARALLEL
   int blk = (DBASE_NUM_TRANS+(MC_PROC*MC_NUM_HOSTS)-1)/(MC_PROC*MC_NUM_HOSTS);
   int lb = get_host_num()*blk;
   int ub = min((get_host_num()+1)*blk, DBASE_NUM_TRANS);
#else
   int blk = DBASE_NUM_TRANS;
   int lb = 0;
   int ub = DBASE_NUM_TRANS;
#endif

   hash_function = get_hash_function(NUM_INSERT,2);
   for(k=3;more;k++){
      seconds(t3);
      if (mflag){
         printf("*****************ITER %d\n",k);
         fprintf(summary,"*****************ITER %d\n",k);
      }
      if (hash_pos[pnum]) delete [] hash_pos[pnum];
      hash_pos[pnum] = new int [(k+1)*hash_function];
      bzero((char *) hash_pos[pnum], (k+1)*hash_function*sizeof(int));
      vist = 1;
      
      //form new candidates
#ifdef PARALLEL
      if (pnum == 0) NUM_ACTUAL_INSERT = apriori_gen(k,pnum);
      if (lsupcnt[pnum]) delete [] lsupcnt[pnum];
      lsupcnt[pnum] = new int [NUM_ACTUAL_INSERT+1];
      bzero((char *) lsupcnt[pnum], ((NUM_ACTUAL_INSERT+1)*sizeof(int)));
#else
      NUM_ACTUAL_INSERT = apriori_gen(k, pnum);
#endif
      if (pnum == 0){
   #ifndef EQ
         Largelist->clearlist();
   #endif
         if (oldCand) delete oldCand; 
      }
      printf("NUM_ACTUAL_INSERT = %d\n", NUM_ACTUAL_INSERT);
      fprintf(summary,"NUM_ACTUAL_INSERT = %d\n", NUM_ACTUAL_INSERT);
      if (more){
         for(i=lb; i < ub;i++){
            make_Itemset(trans, data_set[i].item_list, data_set[i].numitem, data_set[i].tid);
            for (j=0; j < trans->numitems(); j++)
               trans_bitvec[trans->item(j)] = 1;
            subset(trans, 0, trans->numitems()-k+1, trans->numitems(), 
                   Candidate,
                   k, 0, pnum, trans_bitvec, &vist, hash_function);
            for (j=0; j < trans->numitems(); j++)
               trans_bitvec[trans->item(j)] = 0;  
         }
      }
      if (pnum == 0){
         for (i=0; i < DBASE_MAXITEM; i++) hash_indx[i] = -1;
#ifdef PARALLEL
         if (MC_NUM_HOSTS*MC_PROC > 1)
            cnt_ary = mc_reduce(lsupcnt[pnum], NUM_ACTUAL_INSERT, mflag);
         
            //cnt_ary = sum_large_counts(lsupcnt[pnum], pnum, NUM_ACTUAL_INSERT, k);
         else cnt_ary = lsupcnt[pnum];
#else
         cnt_ary = NULL;
#endif

         int ccnntt = 0;
         form_large(Candidate,k, ccnntt, cnt_ary);

         fprintf(summary, "(%d)it = %d ", k, Largelist->numitems());
         printf("(%d)it = %d ", k, Largelist->numitems());
         more = (Largelist->numitems() > 1);
         NUM_INSERT = choose(Largelist->numitems(),2);
#ifdef BALT
         form_hash_indx(get_hash_function(NUM_INSERT,k+1));
#endif
         oldCand = Candidate;
         Candidate = new HashTree(0, get_hash_function(NUM_INSERT,k+1), 
                                  threshold);
      }
      hash_function = get_hash_function(NUM_INSERT,k+1);
      printf("hash_function = %d, NUM_INSERT=%d\n", hash_function, NUM_INSERT);      
      fprintf(summary,"hash_function = %d, NUM_INSERT=%d\n", hash_function, NUM_INSERT);      
      seconds(t4);
      if (pnum == 0) printf(" %f ", t4-t3);
      if (pnum == 0) fprintf(summary," %f ", t4-t3);
   }
   delete [] trans_bitvec;
   delete trans;

   seconds(t2);
   printf("MAIN %f ", t2-t1);
   fprintf(summary,"MAIN %f ", t2-t1);
   return NULL;
}

void init_var()
{
   int i;
   Largelist = new ListItemset;

   start = new int *[NPROC];
   enda = new int *[NPROC];
   hash_pos = new int *[NPROC];
   for (i=0; i < NPROC; i++) hash_pos[i] = NULL;
   Dbase_file = new FILE* [NPROC];

   lsupcnt = new int *[NPROC];
   for (i=0; i < NPROC; i++) lsupcnt[i] = NULL;
   more = 1;
}

void clean_up(){
   int i;

   delete Candidate;
   if (Largelist) delete Largelist;
   delete [] hash_indx;

   for (i=0; i < NPROC; i++){
      delete start[i];
      delete enda[i];
      delete hash_pos[i];
   }
   delete [] start;
   delete [] enda;
   delete [] hash_pos;
   for (i=0; i < NPROC; i++) delete lsupcnt[i];
   delete [] lsupcnt;
}

int main(int argc, char **argv)
{
   char sumfile[100];
   
   parse_args(argc, argv);

   if (mflag) sprintf(sumfile, "apriori.summary.out");
      else sprintf(sumfile, "/dev/null");

   if ((summary = fopen (sumfile, "w")) == NULL){
      printf("can't open %s\n", sumfile);
      exit(-1);
   }

   seconds(ts);
   
#if defined PARALLEL
   if (MC_NUM_HOSTS*MC_PROC > 1){
      MC_SIZE = 16*MC_CHANNEL_SIZE;
      shmeminit(mflag, argc, argv, host_num);
   }
   
#endif
   
   init_var();
 
   fprintf(summary,"%d %d ", MC_PROC, MC_NUM_HOSTS);
   printf( "%d %d ", MC_PROC, MC_NUM_HOSTS);
   
#ifdef PARALLEL 
   if (more) main_proc(NULL);
   if (MC_NUM_HOSTS*MC_PROC > 1) shmemclean();
#else
   if (more) main_proc(NULL);
#endif

   seconds(te);
   printf("Cands %d Total time = %f, IO_time = %f,Comp_time=%f\n", tot_cand, te-ts, t_io_2 - t_io_1, te-ts-t_io_2+t_io_1);
   fprintf(summary, "Cands %d Total time = %f, IO_time = %f,Comp_time=%f\n", tot_cand, te-ts, t_io_2 - t_io_1, te-ts-t_io_2+t_io_1);
  
   fprintf(summary, "\n");
   fflush(summary);
   fflush(stdout);
   clean_up();
   fclose(summary);
   exit(0);
}



























